#!/usr/bin/perl
#
# pftpfxp daemon (c) tanesha team
# bootstrap for pftpfxp using an udp protocol
#
#####################################################################

use IO::Socket::INET;

die "Syntax: $0 <config file>\n" unless @ARGV > 0;

my $PFTPCONNECT = 'cd ~/pf; ./pftp';
my $DAEMONPORT = 4500;
my $PASSWORD = 'secret';

my $config = read_config($ARGV[0]);

my $sock = new IO::Socket::INET(LocalPort => $DAEMONPORT,
				Proto => "udp") || die @!;

printf "Waiting for release (ctrl-c to quit) ..\n";

while ($sock->recv($buf, 1024)) {

    chomp $buf;

    printf("buf = '%s'\n", $buf);

    my ($password, $rel) = split(/ /, $buf);
    printf("rel = $rel, pass = $password\n");

    if ($password ne $PASSWORD) {
	printf "Wrong password from client (%s), ignoring request ..\n", $password;
	next;
    }

    printf "New release .. %s checking if we have setup..\n", $rel;

    do_release($rel);

    printf "Waiting for new release (ctrl-c to quit) ..\n";
}

exit 0;

sub do_release {
    my ($rel) = @_;

    my $dk = $config->{matchers};

    foreach my $dset (@$dk) {

	my @av = split / /, $dset;

	my $m = 0;
	foreach my $k (@av) {
	    $m++ if $rel =~ /$k/i;
	}

	next unless $m == scalar(@av);

	call_pftp($config->{pftpv}->{$dset});

	last;
    }
}

sub call_pftp {
    my ($line) = @_;

    system("$PFTPCONNECT $line");
}

sub read_config {
    my ($fn) = @_;

    my @data = ();
    my @dkeys = ();

    open FH, $fn || die "Cant read config file\n";
    while (<FH>) {

	chomp;
	next if /^\#/;

	my ($k, $v) = $_ =~ /([^=]+)=(.*)/;

	next unless $k && $v;

	if ($k eq 'pftpconnect') {
	    printf("+ setting pftp line: %s\n", $v);
	    $PFTPCONNECT = $v;
	    next;
	}
	elsif ($k eq 'daemonport') {
	    printf("+ setting daemon port: %d\n", $v);
	    $DAEMONPORT = $v;
	    next;
	}
	elsif ($k eq 'password') {
	    printf("+ setting password: %s\n", $v);
	    $PASSWORD = $v;
	    next;
	}

	$k =~ s/ +/ /g;

	$data{$k} = $v;
	push @dkeys, $k;

	printf("+ config -> '%s' -> '%s'\n", $k, $v);
    }

    close FH;

    my %c = (matchers => \@dkeys,
	     pftpv => \%data);

    return \%c;
}				
				
